#!/usr/bin/env python3

# Generates a makefile for a LaTEX document.
#
#   Copyright (C) 2020 Xavier Provençal
#   
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#   
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


from subprocess import getoutput, getstatusoutput
from sys import argv, exit
from os import path

TMP_AND_OUTPUT_FILES_EXT = [ '.aux', '.bbl', '.blg', '.fff', '.idx', 
        '.log', '.log', '.out', '.pdf', '.spl', '.toc', '.tps',
        '.synctex.gz', '.nav', '.snm', '.fdb_latexmk', '.fls'
        ]

def usage():
    print("""Usage : genLatexMakefile [options]

Creates a makefile for a LaTEX document in the current directory with the
following actions :

 make
 Compiles the LaTEX document.

 make view
 Compiles the document if needed and opens the output document with a viewer.

 make edit
 Opens the main source file with a text editor.

 make editall
 Opens all source files with a text editor.

 make clean
 Deletes all temporary files generated by the compilation of the document.


Options:

 -h, -help, --help
 Shows this help and exit.

 --help default
 Shows the default behavior of the script and exit.

 --main TEXFILE   
 The main LaTEX source file to compile. See above for default behavior.

 --outputFormat XXX  
 The extentions of the output document.

 --textEditor CMD
 The command that invokes the text editor.

 --texCompiler CMD
 The command that invokes the LaTEX compiler.

 --viewer CMD
 The command that invokes the document viewer.
    """
    )


DEFAULTARGS = {
        'Linux' : {
            'outputFormat' : 'pdf',
            'textEditor'   : 'gvim -v --servername MAINTEXEDITOR -p',
            'texCompiler'  : 'latexmk -pdf -synctex=1',
            'viewer'       : 'okular 2> /dev/null',
            'main'         : None,
            },
        'Darwin' : {
            'outputFormat' : 'pdf',
            'textEditor'   : 'vim -p',
            'texCompiler'  : 'pdflatex -file-line-error',
            'viewer'       : 'open',
            'main'         : None,
            },
        }

def getDefaultArgs():
    return DEFAULTARGS[ getoutput('uname') ]

def usageDefault():
    print("""DEFAULT BEHAVIOR

If argument `--main` is not specified, the script looks for a file with extention
`.tex.latexmain`. If such a file is found in the current working directory then
the corresponding .tex file is considered as the main source file.

If there is no `.tex.latexmain` file then all `.tex` files are scanned. If only one
of them contains the instruction `\\documentclass{ ... }`, then this file is
considred as the main source.

If a file named `includeonly.tex` is present in the current directory, the
`make edit` will open the files listed in the `\\includeonly{ ... }` command in
addition with the main latex file.

Note: 

The other arguments are system dependant. The system is determined by the
output of the `uname` command.
    """)
    for system in DEFAULTARGS:
        print('    {}'.format(system))
        for arg in DEFAULTARGS[system]:
            lng = max(map(len, DEFAULTARGS[system].keys()))
            if arg != 'main':
                print('        {}{}: `{}`'.format(arg, ' '*(lng-len(arg)+1) ,DEFAULTARGS[system][arg]))
        print('')




def findMainTexFile():
    # test if .tex.latexmain is specified
    s,o = getstatusoutput('ls *.tex.latexmain')
    if s == 0:
        files = o.split('\n')
        if len(files) > 1:
            print('Error, more than one file has the extension .tex.latexmain')
            exit(-1)
        elif len(files) == 1:
            return files[0][:-10]

    # search .tex files for the \documentclass instruction
    s,o = getstatusoutput('grep -l "\\documentclass" *.tex')
    if s == 0:
        files = o.split('\n')
        if len(files) == 0:
            print('Error, no .tex file contains the \\documentclass instruction, use --main to specify a tex file')
            exit(-1)
        elif len(files) > 1:
            print('Error, more than one .tex file containing the \\documentclass instruction.')
            print('Use --main to specify the main source file.')
            exit(-1)
        elif len(files) == 1:
            return files[0]
    print('Error, unable to find the main source file.')
    print('Use --main to specify the main source file.')
    exit(-1)

def readArgs():
    args = getDefaultArgs()
    i = 1
    while i < len(argv):
        if argv[i] in [ '-h', '--help', '-help']:
            if (len(argv) > i+1) and argv[i+1] == "default":
                usageDefault()
                exit(0)
            usage()
            exit(0)
        elif argv[i] == '--outputFormat':
            args['outputFormat'] = argv[i+1]
            i += 1
        elif argv[i] == '--textEditor':
            args['textEditor'] = argv[i+1]
            i += 1
        elif argv[i] == '--texCompiler':
            args['texCompiler'] = argv[i+1]
            i += 1
        elif argv[i] == '--viewer':
            args['viewer'] = argv[i+1]
            i += 1
        elif argv[i] == '--main':
            args['main'] = argv[i+1]
            i += 1
        else:
            usage()
            exit(-1)
        i += 1
    return args

def tempAndOutputFiles(main, included):
    prefix = main[:-4]
    l = []
    l.extend(map(lambda x : prefix + x, TMP_AND_OUTPUT_FILES_EXT))
    for filename in included:
        l.append(filename[:-4] + '.aux')
    return l

def listIncludedFiles(main):
    output = getoutput('grep "\\include{{" {}'.format(main)).split('\n')
    output = list(filter(lambda x : x.find('%') == -1 or x.find('%') > x.find('\\include{'), output))
    included = []
    for o in output:
        starts = o.find('\\include{')
        ends = o.find('}', starts+9)
        while starts != -1 and ends != -1:
            included.append(o[starts+9 : ends] + '.tex')
            starts = o.find('\\include{', ends)
            ends = o.find('}', starts+9)
    return included

def listIncludeOnlyFiles():
    s,_ = getstatusoutput("ls includeonly.tex")
    if s == 0 :
        _,o = getstatusoutput("sed 's/%.*//' includeonly.tex | grep '\\includeonly{.*}' | tail -n 1 | cut -d '{' -f2 | cut -d'}' -f1")
        if not o.isspace() and len(o) > 1:
            def f(name):
                if name[-4:] != '.tex':
                    name += '.tex'
                return name
            return list(map(f, o.split(',')))
    return []

def isBeamer(main):
    texcode = open(main, 'r').readlines()
    l = list(filter(lambda x : (not x.strip().startswith('%')) and x.strip().startswith('\\documentclass') != -1, texcode))
    if len(l) == 0:
        return False
    return l[0].find('beamer') != -1

if __name__ == '__main__':
    args = readArgs()
    if args['main'] == None:
        args['main'] = findMainTexFile()

    texComiler   = args['texCompiler']
    outputFormat = args['outputFormat']
    textEditor   = args['textEditor']
    texCompiler  = args['texCompiler']
    viewer       = args['viewer']
    main         = args['main']
    included     = listIncludedFiles(main)
    includeOnly  = listIncludeOnlyFiles()

    # delete any existing .tex.latexmain file and create the `good` one.
    getoutput('rm *.tex.latexmain')
    getoutput('touch {}.latexmain'.format(main))
    

    print('##########################################')
    print('# TEX compiler  : {}'.format(texComiler))
    print('# Output format : .{}'.format(outputFormat))
    print('# Text editor   : {}'.format(textEditor))
    print('# Output viewer : {}'.format(viewer))
    print('# Main source   : {}'.format(main))
    if len(included) > 0:
        print('# Included      : {}'.format(', '.join(included)))
    if len(includeOnly) > 0:
        print('# IncludeOnly   : {}'.format(', '.join(includeOnly)))
    print('##########################################')

    f = open('makefile', 'w')
    f.writelines([
        'all:\n',
        '\t{} {}\n'.format(texComiler, main),
        '\n\n',
        'edit:\n',
        '\t{} {}\n'.format(textEditor, ' '.join([main]+includeOnly)),
        '\n\n',
        'editall:\n',
        '\t{} {} {}\n'.format(textEditor, main, ' '.join(included)),
        '\n\n',
        'view:\n',
        '\t{} {} &\n'.format(viewer, main[:-3] + outputFormat),
        '\n\n',
        'clean:\n',
        #'\t rm -f ' + ' '.join(tempAndOutputFiles(main, included)) + '\n',
        '\t latexmk -c',
        '\n',
        ])
    if isBeamer(main):
        mainhandout = main[:-4] + '-handout.tex'
        f.writelines([
            '\n\n',
            'handout:\n',
            '\t' + r"sed 's/\]{beamer}/,handout\]{beamer}/' " + main + " > " + mainhandout + '\n',
            '\t{} {}\n'.format(texComiler, mainhandout),
            '\n\n',
            'viewhandout: handout\n',
            '\t{} {} &\n'.format(viewer, mainhandout[:-3] + outputFormat),
            ])
    f.close()
    exit(0)

